const ext = @cImport({
    @cInclude("SDL2/SDL.h");
});

const core = @import("core");
const oar = @import("oar");
const std = @import("std");

///
/// Thread-safe platform abstraction over multiplexing system I/O processing and event handling.
///
pub const App = struct {
    message_chain: ?*Message = null,
    message_semaphore: *ext.SDL_sem,
    message_mutex: *ext.SDL_mutex,
    data: FileSystem,
    user: FileSystem,

    ///
    /// Enqueues `message` to the message chain in `app`.
    ///
    fn enqueue(app: *App, message: *Message) void {
        {
            // TODO: Error check these.
            _ = ext.SDL_LockMutex(app.message_mutex);

            defer _ = ext.SDL_UnlockMutex(app.message_mutex);

            if (app.message_chain) |message_chain| {
                message_chain.next = message;
            } else {
                app.message_chain = message;
            }
        }

        // TODO: Error check this.
        _ = ext.SDL_SemPost(app.message_semaphore);
    }

    ///
    /// Asynchronously executes `procedure` with `arguments` as an anonymous struct of its arguments
    /// and `app` as its execution context.
    ///
    /// Once the execution frame resumes, the value returned by executing `procedure` is returned.
    ///
    pub fn schedule(app: *App, procedure: anytype,
        arguments: anytype) core.meta.FnReturn(@TypeOf(procedure)) {

        const Task = struct {
            procedure: @TypeOf(procedure),
            arguments: *@TypeOf(arguments),
            result: core.meta.FnReturn(@TypeOf(procedure)),

            const Task = @This();

            fn process(userdata: *anyopaque) void {
                const task = @ptrCast(*Task, @alignCast(@alignOf(Task), userdata));

                task.result = @call(.{}, task.procedure, task.arguments.*);
            }
        };

        var task = Task{
            .procedure = procedure,
            .arguments = &arguments,
        };

        var message = Message{
            .kind = .{.task = .{
                .data = &task,
                .action = Task.process,
                .frame = @frame(),
            }},
        };

        suspend app.enqueue(&message);
    }

    ///
    /// Asynchronously logs `info` with `logger` as the logging method and `app` as the execution
    /// context.
    ///
    pub fn log(app: *App, logger: Logger, info: []const u8) void {
        var message = Message{
            .kind = .{.log = .{
                .logger = logger,
                .info = info,
            }},
        };

        app.enqueue(&message);
    }
};

///
/// Platform-agnostic mechanism for working with an abstraction of the underlying file-system(s)
/// available to the application in a sandboxed environment.
///
pub const FileSystem = union(enum) {
    native: []const u8,
    archive: Archive,

    ///
    /// Archive file system information.
    ///
    const Archive = struct {
        file_access: core.io.FileAccess,
        index_cache: IndexCache,
        entry_table: [max_open_entries]Entry = std.mem.zeroes([max_open_entries]Entry),

        ///
        /// Hard limit on the maximum number of entries open at once.
        ///
        const max_open_entries = 16;

        ///
        /// Stateful extension of an [oar.Entry].
        ///
        const Entry = struct {
            owner: ?*core.io.FileAccess,
            cursor: u64,
            header: oar.Entry,
        };

        ///
        /// Table cache for associating [oar.Path] values with offsets to entries in a given file.
        ///
        const IndexCache = core.table.Hashed(oar.Path, u64, .{
            .equals = oar.Path.equals,
            .hash = oar.Path.hash,
        });
    };

    ///
    /// With files typically being backed by a block device, they can produce a variety of errors -
    /// from physical to virtual errors - these are all encapsulated by the API as general
    /// [OpenError.FileNotFound] errors.
    ///
    /// When a given [FileSystem] does not support a specified [OpenMode],
    /// [OpenError.ModeUnsupported] is used to inform the consuming code that another [OpenMode]
    /// should be tried or, if no mode other is suitable, that the resource is effectively
    /// unavailable.
    ///
    /// If the number of known [FileAccess] handles has been exhausted, [OpenError.OutOfFiles] is
    /// used to communicate this.
    ///
    pub const OpenError = error {
        FileNotFound,
        ModeUnsupported,
        OutOfFiles,
    };

    ///
    /// [OpenMode.readonly] indicates that an existing file is opened in a read-only state,
    /// disallowing write access.
    ///
    /// [OpenMode.overwrite] indicates that an empty file has been created or an existing file has
    /// been completely overwritten into.
    ///
    /// [OpenMode.append] indicates that an existing file that has been opened for reading from and
    /// writing to on the end of existing data.
    ///
    pub const OpenMode = enum {
        readonly,
        overwrite,
        append,
    };

    ///
    /// Attempts to open the file identified by `path` with `mode` as the mode for opening the file.
    ///
    /// Returns a [FileAccess] reference that provides access to the file referenced by `path`or a
    /// [OpenError] if it failed.
    ///
    pub fn open(file_system: *FileSystem, path: Path, mode: OpenMode) OpenError!core.io.FileAccess {
        switch (file_system.*) {
            .archive => |*archive| {
                if (mode != .readonly) return error.ModeUnsupported;

                const FileAccess = core.io.FileAccess;

                for (archive.entry_table) |*entry| if (entry.owner == null) {
                    const Implementation = struct {
                        fn archiveEntryCast(context: *anyopaque) *Archive.Entry {
                            return @ptrCast(*Archive.Entry, @alignCast(
                                @alignOf(Archive.Entry), context));
                        }

                        fn close(context: *anyopaque) void {
                            archiveEntryCast(context).owner = null;
                        }

                        fn queryCursor(context: *anyopaque) FileAccess.Error!u64 {
                            const archive_entry = archiveEntryCast(context);

                            if (archive_entry.owner == null) return error.FileInaccessible;

                            return archive_entry.cursor;
                        }

                        fn queryLength(context: *anyopaque) FileAccess.Error!u64 {
                            const archive_entry = archiveEntryCast(context);

                            if (archive_entry.owner == null) return error.FileInaccessible;

                            return archive_entry.header.file_size;
                        }

                        fn read(context: *anyopaque, buffer: []u8) FileAccess.Error!usize {
                            const archive_entry = archiveEntryCast(context);

                            const file_access = archive_entry.owner orelse
                                return error.FileInaccessible;

                            if (archive_entry.cursor >= archive_entry.header.file_size)
                                return error.FileInaccessible;

                            try file_access.seek(archive_entry.header.absolute_offset);

                            return file_access.read(buffer[0 .. std.math.min(
                                buffer.len, archive_entry.header.file_size)]);
                        }

                        fn seek(context: *anyopaque, cursor: usize) FileAccess.Error!void {
                            const archive_entry = archiveEntryCast(context);

                            if (archive_entry.owner == null) return error.FileInaccessible;

                            archive_entry.cursor = cursor;
                        }

                        fn seekToEnd(context: *anyopaque) FileAccess.Error!void {
                            const archive_entry = archiveEntryCast(context);

                            if (archive_entry.owner == null) return error.FileInaccessible;

                            archive_entry.cursor = archive_entry.header.file_size;
                        }

                        fn skip(context: *anyopaque, offset: i64) FileAccess.Error!void {
                            const archive_entry = archiveEntryCast(context);

                            if (archive_entry.owner == null) return error.FileInaccessible;

                            if (offset < 0) {
                                const math = std.math;

                                archive_entry.cursor = math.max(0,
                                    archive_entry.cursor - math.absCast(offset));
                            } else {
                                archive_entry.cursor += @intCast(u64, offset);
                            }
                        }
                    };

                    const Header = oar.Entry;

                    if (archive.index_cache.lookup(path)) |index| {
                        archive.file_access.seek(index) catch return error.FileNotFound;

                        entry.* = .{
                            .owner = &archive.file_access,
                            .cursor = 0,

                            .header = (Header.next(archive.file_access) catch
                                return error.FileNotFound) orelse {

                                // Remove cannot fail if lookup succeeded.
                                std.debug.assert(archive.index_cache.remove(path) != null);

                                return error.FileNotFound;
                            },
                        };
                    } else {
                        while (Header.next(archive.file_access) catch
                            return error.FileNotFound) |entry_header| {

                            if (entry.header.path.equals(path)) entry.* = .{
                                .owner = &archive.file_access,
                                .cursor = 0,
                                .header = entry_header,
                            };
                        }

                        return error.FileNotFound;
                    }

                    return FileAccess{
                        .context = entry,

                        .implementation = &.{
                            .close = Implementation.close,
                            .queryCursor = Implementation.queryCursor,
                            .queryLength = Implementation.queryLength,
                            .read = Implementation.read,
                            .seek = Implementation.seek,
                            .seekToEnd = Implementation.seekToEnd,
                            .skip = Implementation.skip,
                        },
                    };
                };

                return error.OutOfFiles;
            },

            .native => |native| {
                if (native.len == 0) return error.FileNotFound;

                const mem = std.mem;
                var path_buffer = mem.zeroes([4096]u8);
                const seperator_length = @boolToInt(native[native.len - 1] != oar.Path.seperator);

                if ((native.len + seperator_length + path.length) >=
                    path_buffer.len) return error.FileNotFound;

                mem.copy(u8, &path_buffer, native);

                if (seperator_length != 0) path_buffer[native.len] = oar.Path.seperator;

                mem.copy(u8, path_buffer[native.len .. path_buffer.
                    len], path.buffer[0 .. path.length]);

                const FileAccess = core.io.FileAccess;

                const Implementation = struct {
                    fn rwOpsCast(context: *anyopaque) *ext.SDL_RWops {
                        return @ptrCast(*ext.SDL_RWops, @alignCast(
                            @alignOf(ext.SDL_RWops), context));
                    }

                    fn close(context: *anyopaque) void {
                        ext.SDL_ClearError();

                        if (ext.SDL_RWclose(rwOpsCast(context)) != 0)
                            ext.SDL_LogWarn(ext.SDL_LOG_CATEGORY_APPLICATION, ext.SDL_GetError());
                    }

                    fn queryCursor(context: *anyopaque) FileAccess.Error!u64 {
                        ext.SDL_ClearError();

                        const sought = ext.SDL_RWtell(rwOpsCast(context));

                        if (sought < 0) return error.FileInaccessible;

                        return @intCast(u64, sought);
                    }

                    fn queryLength(context: *anyopaque) FileAccess.Error!u64 {
                        ext.SDL_ClearError();

                        const sought = ext.SDL_RWsize(rwOpsCast(context));

                        if (sought < 0) return error.FileInaccessible;

                        return @intCast(u64, sought);
                    }

                    fn read(context: *anyopaque, buffer: []u8) FileAccess.Error!usize {
                        ext.SDL_ClearError();

                        const buffer_read = ext.SDL_RWread(rwOpsCast(
                            context), buffer.ptr, @sizeOf(u8), buffer.len);

                        if ((buffer_read == 0) and (ext.SDL_GetError() != null))
                            return error.FileInaccessible;

                        return buffer_read;
                    }

                    fn seek(context: *anyopaque, cursor: usize) FileAccess.Error!void {
                        var to_seek = cursor;

                        while (to_seek != 0) {
                            const math = std.math;
                            const sought = math.min(to_seek, math.maxInt(i64));

                            ext.SDL_ClearError();

                            if (ext.SDL_RWseek(rwOpsCast(context), @intCast(i64, sought),
                                ext.RW_SEEK_CUR) < 0) return error.FileInaccessible;

                            to_seek -= sought;
                        }
                    }

                    fn seekToEnd(context: *anyopaque) FileAccess.Error!void {
                        ext.SDL_ClearError();

                        if (ext.SDL_RWseek(rwOpsCast(context), 0, ext.RW_SEEK_END) < 0)
                            return error.FileInaccessible;
                    }

                    fn skip(context: *anyopaque, offset: i64) FileAccess.Error!void {
                        ext.SDL_ClearError();

                        if (ext.SDL_RWseek(rwOpsCast(context), offset, ext.RW_SEEK_SET) < 0)
                            return error.FileInaccessible;
                    }
                };

                ext.SDL_ClearError();

                return FileAccess{
                    .context = ext.SDL_RWFromFile(&path_buffer, switch (mode) {
                        .readonly => "rb",
                        .overwrite => "wb",
                        .append => "ab",
                    }) orelse return error.FileNotFound,

                    .implementation = &.{
                        .close = Implementation.close,
                        .queryCursor = Implementation.queryCursor,
                        .queryLength = Implementation.queryLength,
                        .read = Implementation.read,
                        .seek = Implementation.seek,
                        .seekToEnd = Implementation.seekToEnd,
                        .skip = Implementation.skip,
                    },
                };
            },
        }
    }
};

///
///
///
pub const Graphics = opaque {
    ///
    ///
    ///
    pub const Event = struct {
        keys_up: Keys = std.mem.zeroes(Keys),
        keys_down: Keys = std.mem.zeroes(Keys),
        keys_held: Keys = std.mem.zeroes(Keys),

        const Keys = [256]bool;
    };

    ///
    ///
    ///
    const Implementation = struct {
        event: Event,
    };

    ///
    ///
    ///
    pub fn poll(graphics: *Graphics) ?*const Event {
        _ = graphics;

        return null;
    }

    ///
    ///
    ///
    pub fn present(graphics: *Graphics) void {
        // TODO: Implement;
        _ = graphics;
    }
};

///
/// [Logger.info] logs information that isn't necessarily an error but indicates something useful to
/// be logged.
///
/// [Logger.debug] logs information only when the engine is in debug mode.
///
/// [Logger.warning] logs information to indicate a non-critical error has occured.
///
pub const Logger = enum(u32) {
    info = ext.SDL_LOG_PRIORITY_INFO,
    debug = ext.SDL_LOG_PRIORITY_DEBUG,
    warning = ext.SDL_LOG_PRIORITY_WARN,
};

///
/// Linked list of asynchronous messages chained together to be processed by the work processor.
///
pub const Message = struct {
    next: ?*Message = null,

    kind: union(enum) {
        quit,

        log: struct {
            logger: Logger,
            info: []const u8,
        },

        task: struct {
            data: *anyopaque,
            action: fn (*anyopaque) void,
            frame: anyframe,
        },
    },
};

///
/// Path to a file on a [FileSystem].
///
pub const Path = oar.Path;

///
/// [RunError.InitFailure] occurs when the runtime fails to initialize.
///
pub const RunError = error {
    InitFailure,
};

///
/// Runs a graphical application referenced by `run` with `error` as its error set.
///
/// Should an error from `run` occur, an `Error` is returned, otherwise a [RunError] is returned if
/// the underlying runtime fails and is logged.
///
pub fn display(comptime Error: anytype,
    comptime run: fn (*App, *Graphics) callconv(.Async) Error!void) (RunError || Error)!void {

    var cwd = FileSystem{.native = "./"};
    const user_prefix = ext.SDL_GetPrefPath("ona", "ona") orelse return error.InitFailure;

    defer ext.SDL_free(user_prefix);

    var gpa = std.heap.GeneralPurposeAllocator(.{}){};

    defer if (gpa.deinit())
        ext.SDL_LogWarn(ext.SDL_LOG_CATEGORY_APPLICATION, "Runtime allocator leaked memory");

    var app = App{
        .user = .{.native = std.mem.sliceTo(user_prefix, 0)},

        .message_semaphore = ext.SDL_CreateSemaphore(0) orelse {
            ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION, "Failed to create message semaphore");

            return error.InitFailure;
        },

        .message_mutex = ext.SDL_CreateMutex() orelse {
            ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION, "Failed to create message mutex");

            return error.InitFailure;
        },

        .data = .{.archive = .{
            .index_cache = FileSystem.Archive.IndexCache.init(gpa.allocator()) catch
                return error.InitFailure,

            .file_access = cwd.open(try Path.joined(&.{"./data.oar"}), .readonly) catch {
                ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION, "Failed to load ./data.oar");

                return error.InitFailure;
            },
        }},
    };

    defer {
        app.data.archive.file_access.close();
        app.data.archive.index_cache.deinit();
        ext.SDL_DestroySemaphore(app.message_semaphore);
        ext.SDL_DestroyMutex(app.message_mutex);
    }

    const message_thread = ext.SDL_CreateThread(processMessages, "Message Processor", &app) orelse {
        ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION, "Failed to create message processor");

        return error.InitFailure;
    };

    defer {
        var message = Message{.kind = .quit};

        app.enqueue(&message);

        {
            var status = std.mem.zeroes(c_int);

            // SDL2 defines waiting on a null thread reference as a no-op. See
            // https://wiki.libsdl.org/SDL_WaitThread for more information
            ext.SDL_WaitThread(message_thread, &status);

            if (status != 0) {
                // TODO: Error check this.
            }
        }
    }

    if (ext.SDL_Init(ext.SDL_INIT_EVERYTHING) != 0) {
        ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION, "Failed to initialize runtime");

        return error.InitFailure;
    }

    defer ext.SDL_Quit();

    const window = create_window: {
        const pos = ext.SDL_WINDOWPOS_UNDEFINED;
        var flags = @as(u32, 0);

        break: create_window ext.SDL_CreateWindow("Ona", pos, pos, 640, 480, flags) orelse {
            ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION, "Failed to create window");

            return error.InitFailure;
        };
    };

    defer ext.SDL_DestroyWindow(window);

    const renderer = create_renderer: {
        var flags = @as(u32, 0);

        break: create_renderer ext.SDL_CreateRenderer(window, -1, flags) orelse {
            ext.SDL_LogCritical(ext.SDL_LOG_CATEGORY_APPLICATION, "Failed to create renderer");

            return error.InitFailure;
        };
    };

    defer ext.SDL_DestroyRenderer(renderer);

    var graphics = Graphics.Implementation{
        .event = .{

        },
    };

    return run(@ptrCast(*App, &app), @ptrCast(*Graphics, &graphics));
}

///
/// [FileSystemMessage] processing function used by a dedicated worker thread, where `data`
/// is a type-erased reference to a [EventLoop].
///
/// The processor returns `0` if it exited normally or any other value if an erroneous exit
/// occured.
///
pub fn processMessages(userdata: ?*anyopaque) callconv(.C) c_int {
    const app = @ptrCast(*App, @alignCast(@alignOf(App), userdata orelse unreachable));

    while (true) {
        // TODO: Error check these.
        _ = ext.SDL_SemWait(app.message_semaphore);
        _ = ext.SDL_LockMutex(app.message_mutex);

        defer _ = ext.SDL_UnlockMutex(app.message_mutex);

        while (app.message_chain) |message| {
            switch (message.kind) {
                .quit => return 0,

                .log => |log| ext.SDL_LogMessage(ext.SDL_LOG_CATEGORY_APPLICATION,
                    @enumToInt(log.logger), "%.*s", log.info.len, log.info.ptr),

                .task => |task| {
                    task.action(task.data);

                    resume task.frame;
                },
            }

            app.message_chain = message.next;
        }
    }
}
